"""
Module defining the TimeSeries object
"""

from typing import IO, Union, Generic, TypedDict, Sequence, List, Callable, Type, Optional, Iterable, Dict, Tuple
from pathlib import Path
import warnings

import io 
import csv
import os
import codecs

from dataclasses import dataclass
import numpy as np

from datetime import datetime
from dateutil import parser

from .datetimelikearray import DatetimeLikeArray, TimeDeltaLike, DatetimeLike

# Default times_conversion function
def parse_time(s: str):
    try: 
        n = float(s)
        return n
    except ValueError:
        return parser.parse(s)

class ShapeChangedWarning(Warning):
    """
    Warning for cases where `TimeSeries.__post_init__` alters the shape of the dependent variable.
    """


class TimeSeriesTypedDict(TypedDict):
    """
    TypedDict form of a TimeSeries object, useful for turning into json
    """

    dependent_variable: Sequence[Sequence[float]]
    times: Sequence[Union[float, str]]


@dataclass
class TimeSeries(Generic[TimeDeltaLike]):
    """
    A class representing a time series, encapsulating dependent variables and corresponding timestamps.
    """

    dependent_variable: np.typing.NDArray[np.floating]
    """Array of dependent variables representing the time series values."""

    times: DatetimeLikeArray
    """Array of time values associated with the dependent variable."""
    
    def __post_init__(self):
        """
        Ensures proper formatting of time values and checks for uniform spacing.
        Raises an error if the time values are not uniformly spaced.
        """
        # If List[float | int | datetime] is passed, convert to NDArray, NDArray, or DatetimeLikeArray respectively
        if isinstance(self.times[0], datetime):
            # For some reason, NumPy does not automatically handle datetime object dtypes as np.datetime64.
            dtype_ = 'datetime64[s]'
        else:
            dtype_ = np.dtype(type(self.times[0]))
        
        self.times = DatetimeLikeArray(self.times, dtype=dtype_)
        timesteps = np.diff(self.times)
        
        if not np.isclose(np.std(timesteps.astype(float)), 0.0):
            raise ValueError("TimeSeries.times must be uniformly spaced")
        if np.isclose(np.mean(timesteps.astype(float)), 0.0):
            raise ValueError("TimeSeries.times must have a timestep greater than zero")
        
        if not isinstance(self.dependent_variable, np.ndarray):
            self.dependent_variable = np.array(self.dependent_variable)
        
        if len(self.dependent_variable.shape) == 1:
            num_steps = len(self.dependent_variable)
            self.dependent_variable = self.dependent_variable.reshape(num_steps, 1)
            warnings.warn(
                f"TimeSeries.dependent_variable should have shape (number of steps, dimensionality). "
                f"The shape has been changed from {(num_steps,)} to {self.dependent_variable.shape}",
            )
    
    def save(self, fp: Union[IO, str, Path], header: str = "", delimiter=",", fmt: Union[str, Tuple[str]] = '%s'):
        """
        Saves the time series data to a file.
        
        Args:
            fp (Union[IO, str, Path]): File path or object where the TimeSeries instance will be saved.
            header (str, optional): An optional header. Defaults to an empty string.
            delimiter (str, optional): The delimiter used in the output file. Defaults to a comma.
            fmt (Tuple(str), optional): Format specifier used for saving the data to fp. Defaults to '%s' (String representation)
        """
        
        data = np.vstack((
            self.times.astype('object'), 
            self.dependent_variable.T.astype('object')
        ), dtype='object').T
        
        np.savetxt(
            fp,
            data,
            delimiter=delimiter,
            fmt=fmt,
            header=header,
            comments=""
        )
    
    @property
    def num_dims(self) -> int:
        """
        Returns the dimensionality of the dependent variable.
        
        Returns:
            int: Number of dimensions of the time series.
        """
        return self.dependent_variable.shape[1]
    
    @classmethod
    def from_csv(
        cls, 
        fp: Union[IO, str, Path, str], 
        time_col: str,
        times_dtype: Type,
        dependent_var_cols: List[str],
        times_conversion: Callable[[str], DatetimeLike] = parse_time, 
        dep_var_conversion: Callable[[str], float] = float,
        max_rows: Optional[int] = 0
        ):
        """
        Loads time series data from a CSV file.
        
        Args:
            fp (Union[IO, str, Path]): File path or object to read from.
            time_col (str): Name of the times column.
            times_dtype (dtype): Type of the times column
            dependent_var_cols (List[str]): Names of the dependent variable columns
            times_conversion (Callable[[str], DatetimeLike]): Function determining how to parse elements of the times column. Defaults to parse_time
            dep_var_conversion (Callable[[str], float]): Function determining how to parse elements of the dependent variable columns. Defaults to float
            max_rows (float): Maximum number of rows that should be parsed. If set to zero, all rows are parsed. Defaults to 0
        
        Returns:
            TimeSeries: A TimeSeries instance populated by data from the csv file.
        """
        # Helper generator to get a csv.DictReader source from different sources
        def get_reader(fp) -> Iterable[Dict]:
            
            # Checking the input format
            if isinstance(fp, Path) and os.path.isfile(fp):
                with open(fp, encoding='utf-8') as file:
                    yield from csv.DictReader(file, delimiter=",")
        
            elif isinstance(fp, io.BytesIO):
                # Use codecs, because decoding directly is an eager operation
                fp.seek(0) 
                yield from csv.DictReader(codecs.getreader("utf-8")(fp))
            
            elif isinstance(fp, io.StringIO):
                fp.seek(0)
                yield from csv.DictReader(fp, delimiter=",")
            
            else: 
                raise FileNotFoundError()
        
        # Generator for lazy dependent variable processing
        def process_dep(): 
            reader = get_reader(fp)
            for i, row in enumerate(reader):
                if max_rows and i >= max_rows:
                    break
                for col in dependent_var_cols: 
                    yield dep_var_conversion(row[col]) 
        
        # Generator for lazy times processing
        def process_times(): 
            reader = get_reader(fp)
            for i, row in enumerate(reader):
                if max_rows and i >= max_rows:
                    break 
                yield times_conversion(row[time_col])
        
        dep = np.fromiter(process_dep(), dtype=float).reshape(-1, len(dependent_var_cols), order='C')
        
        t = DatetimeLikeArray.from_iter(process_times(), dtype=times_dtype)
        
        return cls(
            dependent_variable=dep, 
            times=t
        )
    
    def to_dict(self) -> TimeSeriesTypedDict:
        
        """
        convert to a typed dictionary
        """

        return TimeSeriesTypedDict(
            dependent_variable=self.dependent_variable.tolist(),
            times=self.times.to_list()
        )
    
    @property
    def timestep(self) -> TimeDeltaLike:
        """
        Returns the time step between consecutive observations.

        Returns:
            TimeDeltaLike: The timestep of the time series.
        """
        return self.times[1] - self.times[0]

    def __eq__(self, other) -> bool:
        """
        Checks equality between two TimeSeries instances.
        
        
        Returns:
            bool: True if both instances have the same times and dependent variables.
        """
        return self.times == other.times and bool(np.all(np.isclose(self.dependent_variable, other.dependent_variable)))

    def __getitem__(self, key):
        """
        Retrieves a subset of the TimeSeries.
        """
        return TimeSeries(self.dependent_variable[key], self.times[key])

    def __setitem__(self, key, value):
        """
        Sets a subset of the TimeSeries.
        """
        if not isinstance(value, TimeSeries):
            raise TypeError("Value must be a TimeSeries object")
        if isinstance(key, slice) and key.stop > len(self.dependent_variable):
            raise ValueError("Slice stop index out of range")
        if isinstance(key, int) and key >= len(self.dependent_variable):
            raise ValueError("Index out of range")
        self.dependent_variable[key] = value.dependent_variable
        self.times[key] = value.times

    def __add__(self, other):
        """
        Adds two TimeSeries instances element-wise.
        """
        if not len(self.times) == len(other.times):
            raise ValueError("Can only add TimeSeries instances that have the same number of timesteps")
        if not np.all(self.times == other.times):
            raise ValueError("Can only add TimeSeries instances that span the same times")
        return TimeSeries(
            dependent_variable=self.dependent_variable + other.dependent_variable,
            times=self.times
        )
    
    
    def __sub__(self, other):
        """
        Subtracts two TimeSeries instances element-wise.
        """
        if not len(self.times) == len(other.times):
            raise ValueError("Can only subtract TimeSeries instances that have the same number of timesteps")
        if not np.all(self.times == other.times):
            raise ValueError("Can only subtract TimeSeries instances that span the same times")
        return TimeSeries(
            dependent_variable=self.dependent_variable - other.dependent_variable,
            times=self.times
        )

    def __rshift__(self, other):
        """
        Concatenates two non-overlapping TimeSeries instances.
        """
        if self.times[-1] >= other.times[0]:
            print(self.times[-1], other.times[0])
            raise ValueError("Can only concatenate TimeSeries instances with non-overlapping time values")
        return TimeSeries(
            dependent_variable=np.vstack((self.dependent_variable, other.dependent_variable)),
            times=np.concatenate((self.times, other.times))
        )

    def __len__(self):
        """
        Returns the number of timesteps in the TimeSeries.
        """
        return len(self.times)



r"""
.. include:: ../README.md

# Examples

## 🐇 Basic Lotka-Volterra example

Below is an example of using `aqua-blue` to predict the predator-prey
[Lotka-Volterra equations](https://en.wikipedia.org/wiki/Lotka%E2%80%93Volterra_equations):

$$ \dot x = \alpha x - \beta xy $$
$$ \dot y = -\gamma y + \delta xy$$

with parameters $\alpha = 0.1$, $\beta = 0.02$, $\gamma = 0.3$, and $\delta = 0.01$, and initial conditions
$(x_0, y_0) = (20, 9)$. We train a reservoir computer with a reservoir dimensionality of $1000$ over $0\leq t\leq 10$,
with $1000$ timesteps. Then, we predict the next $1000$ timesteps.

Here, we use `scipy.integrate.solve_ivp` to integrate the system of differential equations.


```py
.. include:: ../examples/lotka-volterra.py
```

## 🕓 Using datetime objects

Below is an example of a simple sine-cosine task similar to above, using `datetime.datetime` objects as times.

```py
.. include:: ../examples/sine-cosine.py
```

## 📡 Load and output a JSON string

Below is an example of inputting a `json` string as the training data, and outputting a `json` string for the
prediction. This is particularly useful for interfacing `aqua-blue` with already-existing systems.

```py
.. include:: ../examples/json-example.py
```

## 🏋 Explicit weights

Below is an example of generating explicit matrices for $W_\text{in}$ and $W_\text{res}$. Here, `sparsity=0.99` and
`spectral_radius=1.2` respectively zero-out $99\%$ of $W_\text{res}$'s elements and force $W_\text{res}$ to have a
[spectral radius](https://en.wikipedia.org/wiki/Spectral_radius) of $1.2$. We also showcase the `>>` operator, which
concatenates instances of `aqua_blue.time_series.TimeSeries`.

```py
.. include:: ../examples/explicit-weights.py
```

## 📈 Explicit activation function

Below is an example of using a different activation function to map from the input state to the reservoir.
Here, we use both hyperbolic tangent (`tanh`) and the [error function](https://en.wikipedia.org/wiki/Error_function)
(`erf`), and compare the results.

```py
.. include:: ../examples/activation-functions.py
```

## 🔗 HTTP Requests

Below is an example of pulling csv file data from a resource URL using the [requests](https://pypi.org/project/requests/) library. 
Here, we retrieve a time series of temperature data from [NCEI NOAA](https://www.ncei.noaa.gov/data/global-summary-of-the-day/access/2024/01001099999.csv)
and use it for training and predicting temperatures.

```py
.. include:: ../examples/apireq-example.py
```

## 📁 Read from and write to CSV files

Below is an example of parsing data from a csv file (`goldstocks.csv`) and writing it to a `TimeSeries` object, which is used for training and predictions. 
The predictions are written to a new csv file (`predicted-goldstocks.csv`).

```py
.. include:: ../examples/csv-example.py
```
"""

__version__ = "0.2.15"
__authors__ = [
    "Jacob Jeffries",
    "Hrishikesh Belagali",
    "Avik Thumati",
    "Ameen Mahmood",
    "Samuel Josephs",
]
__author_emails__ = [
    "jacob.jeffries@ccmc.global",
    "hrishikesh.belagali@ccmc.global",
    "avik.thumati@ccmc.global",
    "ameen.mahmood@ccmc.global",
    "samuel.josephs@ccmc.global"
]
__url__ = "https://github.com/Chicago-Club-Management-Company/aqua-blue"

from . import utilities as utilities
from . import reservoirs as reservoirs
from . import readouts as readouts
from . import models as models
from . import time_series as time_series



"""
Module defining readout layers.

This module provides the abstract `Readout` class and its concrete implementation,
`LinearReadout`. Readout layers map the internal reservoir states of an Echo State
Network (ESN) to output values.

Classes:
    - Readout: Abstract base class defining the interface for readout layers.
    - LinearReadout: A linear mapping readout layer that transforms reservoir states
      into output values using learned coefficients.
"""

from dataclasses import dataclass, field
from abc import ABC, abstractmethod

import numpy as np


@dataclass
class Readout(ABC):
    """
    Abstract base class for readout layers in Echo State Networks (ESNs).

    Readout layers transform the high-dimensional reservoir states into output
    predictions. The transformation is typically learned during training.

    Attributes:
        coefficients (np.ndarray):
            The learned weights for mapping reservoir states to output values.
            This is set after training.
    """

    coefficients: np.typing.NDArray[np.floating] = field(init=False)
    """The learned weight matrix for the readout layer, initialized during training."""

    @abstractmethod
    def train(
        self,
        independent_variables: np.typing.NDArray[np.floating],
        dependent_variables: np.typing.NDArray[np.floating]
    ):
        """
        Trains the readout layer by learning the mapping from reservoir states to output values.

        This method takes independent input variables (reservoir states) and corresponding
        dependent variables (target outputs) to compute the optimal readout weights.

        Args:
            independent_variables (np.ndarray):
                The reservoir state matrix used as input for training.
            dependent_variables (np.ndarray):
                The expected output values corresponding to the input states.
        """

        pass

    @abstractmethod
    def reservoir_to_output(self, reservoir_state: np.typing.NDArray[np.floating]) -> np.typing.NDArray[np.floating]:

        """
        Maps a given reservoir state to an output value.

        Args:
            reservoir_state (np.ndarray):
                The current state of the reservoir.

        Returns:
            np.ndarray: The predicted output corresponding to the given reservoir state.
        """

        pass


@dataclass
class LinearReadout(Readout):
    """
    A linear readout layer that applies a learned linear transformation to reservoir states.

    This readout layer learns a set of coefficients during training and applies a
    simple linear mapping to transform reservoir states into output predictions.

    Attributes:
        rcond (float):
            A regularization parameter used in the pseudo-inverse calculation to
            prevent numerical instability in the least squares solution.
    """

    rcond: float = 1.0e-10
    """
    Regularization parameter for pseudo-inverse computation.

    This controls the minimum singular value considered for the pseudo-inverse 
    computation. A lower value ensures more stable training.
    """

    def train(
        self,
        independent_variables: np.typing.NDArray[np.floating],
        dependent_variables: np.typing.NDArray[np.floating]
    ):
        r"""
        Trains the linear readout layer by solving the least-squares optimization problem.

        The training process determines the optimal readout coefficients $W^*$ by solving the optimization problem
        below:

        $$W^* = \lim_{\lambda\to 0^+} \arg\min_W \| XW - Y\|_F^2 + \lambda \|W\|_F^2$$

        where $X$ is the matrix of reservoir states (independent variables)
        and $Y$ is the matrix of target output values (dependent variables).

        Args:
            independent_variables (np.ndarray):
                The reservoir state matrix used for training.
            dependent_variables (np.ndarray):
                The target output values corresponding to the reservoir states.
        """
        coeff = np.linalg.pinv(independent_variables, rcond=self.rcond) @ dependent_variables
        self.coefficients = coeff.T

    def reservoir_to_output(self, reservoir_state: np.typing.NDArray[np.floating]) -> np.typing.NDArray[np.floating]:
        """
        Computes the output from a given reservoir state using a learned linear mapping.

        This method applies the learned weight matrix (`self.coefficients`) to map
        the reservoir state to an output value.

        Args:
            reservoir_state (np.ndarray):
                The reservoir state to be mapped to an output value.

        Returns:
            np.ndarray: The predicted output value.

        Raises:
            ValueError: If the readout layer has not been trained (i.e., coefficients are not set).
        """
        if not hasattr(self, "coefficients"):
            raise ValueError("Need to train readout before using it")

        return self.coefficients @ reservoir_state
